#pragma once

#include <pwd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/stat.h>
#include <unistd.h>

#include <cstring>
#include <string>
#include <vector>
#include <chrono>
#include <thread>
#include <fstream>
#include <sys/wait.h>
#include "../except.hpp"
#include "../log.hpp"
#include "../process.hpp"

extern char **environ;

namespace fcf {
namespace details {
namespace {

  struct subprocess_params_t {
    subprocess_params_t(const char* a_name, std::vector<std::string> a_argv, std::vector<std::pair<std::string, std::string> > a_envv)
    {
      argv.push_back(std::strcpy(new char[strlen(a_name)+1], a_name));
      for(size_t i = 0; i < a_argv.size(); ++i)
        argv.push_back(std::strcpy(new char[a_argv[i].length() + 1], a_argv[i].c_str()));
      argv.push_back((char*)0);

      for(size_t i = 0; i < a_envv.size(); ++i){
        std::string val = a_envv[i].first + "=" + a_envv[i].second;
        envv.push_back(std::strcpy(new char[val.length() + 1], val.c_str()));
      }
      envv.push_back((char*)0);
    }

    ~subprocess_params_t() {
      for(size_t i = 0; i < argv.size()-1; ++i)
        delete[] argv[i];
      for(size_t i = 0; i < envv.size()-1; ++i)
        delete[] envv[i];
    }

    std::vector<char*> argv;
    std::vector<char*> envv;
  };



  inline pid_t create_process(const char* a_name,
                              std::vector<std::string> a_argv,
                              std::vector<std::pair<std::string, std::string> > a_envv,
                              const char* a_user,
                              const char* a_password) {
    a_name = a_name ? a_name : "";
    a_user = a_user ? a_user : "";
    a_password = a_password ? a_password : "";

    subprocess_params_t params(a_name, a_argv, a_envv);
    passwd* userinfo = 0;
    if (a_user && a_user[0]) {
      userinfo = getpwnam(a_user);
      if (!userinfo)
        throw fcf::except::nouserinfo(a_user);
    }

    struct stat stproc;
    std::string cmdline;
    if (a_name[0])
      cmdline += std::string() + "\"" + a_name + "\"";

    for(size_t i = 0; i < a_argv.size(); ++i){
      if (i != 0 || (i == 0 && a_name[0]))
        cmdline += " ";
      cmdline +=  a_argv[i];
    }

    pid_t pid = fork();
    if (!pid) {
      if (!userinfo || setgid(userinfo->pw_gid) == -1)
        exit(90);
      if (!userinfo || setuid(userinfo->pw_uid) == -1)
        exit(90);
      execve(a_name, &params.argv[0], &params.envv[0]);
      exit(1);
    }

    if (pid == -1 || !pid)
      throw fcf::except::process_start(cmdline.c_str());

    return pid;
  }  

  inline void close_process(pid_t a_pid) {
    if (!a_pid)
      return;
    ::kill(a_pid, SIGKILL);
    ::kill(a_pid, SIGTERM);
  }

  inline bool check_pid(pid_t a_pid) {
    struct stat sts;
    return !(stat((std::string("/proc/") + std::to_string(a_pid)).c_str(), &sts) == -1 && errno == ENOENT);
  }

  inline size_t get_process_memory_size(int a_pid) {
    std::string buff;
    std::ifstream fs(std::string("/proc/") + std::to_string(a_pid) + "/stat");
    std::getline(fs, buff);
    size_t cn(0), pos(0), lstpos(0), l(buff.length());
    for (size_t i = 0; i < l; ++i) {
      if (buff[i] == ' ') {
        lstpos = pos;
        pos = i + 1;
        ++cn;
      }
      if (cn == 24)
        return std::stoul(buff.substr(lstpos, i - lstpos)) * getpagesize();
    }
    return 0;
  }


}
}
}
