#include <climits>
#include <string>
#include <iostream>
#include <sstream>

#define BOOST_BIND_GLOBAL_PLACEHOLDERS
#include <boost/bind.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

#include <fcf/string.hpp>
#include <fcf/log.hpp>

#include "server.hpp"

namespace application {

  server_t::server_t(const server_configuration_t& a_config)
    : _stop(false), _config(a_config) {
    for(int i = 0; i < a_config.endpoints.size(); ++i){
      _endpoints.push_back( endpoint_sptr_t(new endpoint_t(a_config, i)) );
    }
  }

  void server_t::run() {
    fcf::log << "Start listening control port " << _config.serverControlPort;
    _socket.reset(fcf::net::create_socket("127.0.0.1", _config.serverControlPort));
    _socket->set_reuse(true);
    _socket->bind();
    _socket->listen(16);
    _thread = std::thread(&server_t::_handler, this);

    for(int i = 0; i < _endpoints.size(); ++i)
      _endpoints[i]->run();
  }

  void server_t::stop() {
    fcf::log << "Stop server ";
    fcf::log << "--------------------------------------------";
    {
      std::unique_lock<std::mutex> lock(_mutex);
      _stop = true;
    }
    _socket->close();

    if (_thread.joinable())
      _thread.join();

    for(int i = 0; i < _endpoints.size(); ++i)
      _endpoints[i]->stop();
  }

  void server_t::_handler(){
    char buffer[1024*4];

    std::unique_lock<std::mutex> lock(_mutex);
    while(!_stop){
      try {
        lock.unlock();
        fcf::net::socket_sptr_t client;
        int rc;
        try {
          client.reset(_socket->accept());
          rc = client->recv(&buffer[0], sizeof(buffer));
          if (!rc)
            throw new std::runtime_error("Empty inner command");
        } catch(const std::exception& e) {
          lock.lock();
          throw std::runtime_error(e.what());
        }
        lock.lock();
        if (_stop)
          break;
        int pos = fcf::find_str(buffer, 0, rc, "\n");
        if (pos == -1)
          continue;
        std::stringstream ss(std::string(&buffer[pos+1], &buffer[rc]));
        boost::property_tree::ptree ptree;
        boost::property_tree::read_json(ss, ptree);
        if (ptree.get<std::string>("name") == "restart"){
          int         group = ptree.get<int>("group");
          std::string cause = ptree.get<std::string>("cause");
          std::string file;
          if (cause == "file_update")
            file = ptree.get<std::string>("file");
          for(int i = 0; i < _endpoints.size(); ++i){
            if (_endpoints[i]->is_group_exist(group)){
              if (cause == "file_update")
                fcf::log << "Restarting the process due to updating the '" << file << "' file";
              else
                fcf::log << "Restarting processes at the request of the handler for group " << group;
              _endpoints[i]->restart_group(group);
              break;
            }
          }
        }
        client->close();
        client.reset();
      } catch(const std::exception& e) {
        if (!_stop)
          fcf::err << "Control socket error: " << e.what();
      }
    }
  }

}
