#pragma once
#include <atomic>
#include <vector>
#include <string>
#include <thread>
#include <mutex>
#include <condition_variable>

namespace application { 

  #ifdef WIN32
  typedef int pid_t;
  #else
  typedef ::pid_t pid_t;
  #endif

  enum SUBPROCESS_LEAK_PROTECTED {
    SLP_YES,
    SLP_NO,
    SLP_NO_MATTER
  };

  struct request_processing_information_t {
    bool  memory_leak_protection  = true;
    int   client_socket_timeout   = -1;
    int   max_request_timeout     = -1;
    int   server_socket_timeout   = -1;
    int   max_response_timeout    = -1;
  };

  class subprocesses_t {
  public:
    struct subprocesses_state_t {
      std::string                                       name;
      std::vector<std::string>                          argv;
      std::vector<std::pair<std::string, std::string> > envv;
      std::vector<std::string>                          node_options;
      std::string                                       user;
      std::string                                       password;
      int                                               port;
      int                                               control_port;
      int                                               start_timeout;
      int                                               stop_timeout;
      pid_t                                             pid;
      pid_t                                             subpid;
      std::atomic_int32_t                               counter;
      bool                                              active;
      bool                                              leak_protected;

      inline subprocesses_state_t()
        : pid(0), subpid(0) { }

      inline subprocesses_state_t(
        const char*                                                                         a_name, 
        std::vector<std::string>                                                            a_argv,
        std::vector<std::pair<std::string, std::string> >                                   a_envv,
        std::vector<std::string>                                                            a_node_options,
        const char*                                                                         a_user,
        const char*                                                                         a_password,
        int                                                                                 a_port,
        int                                                                                 a_control_port,
        int                                                                                 a_start_timeout,
        int                                                                                 a_stop_timeout,
        pid_t                                                                               a_pid,
        std::atomic_int32_t                                                                 a_counter,
        bool                                                                                a_active,
        bool                                                                                a_leak_protected)
      : name(a_name), argv(a_argv), envv(a_envv), node_options(a_node_options), 
        user(a_user), password(a_password), port(a_port), control_port(a_control_port), 
        start_timeout(a_start_timeout), stop_timeout(a_stop_timeout), pid(a_pid), 
        counter(a_counter.load()), active(a_active), leak_protected(a_leak_protected)
      { }

      inline subprocesses_state_t(const subprocesses_state_t& a_struct)
        : name(a_struct.name), argv(a_struct.argv), envv(a_struct.envv), node_options(a_struct.node_options),
          user(a_struct.user), password(a_struct.password), port(a_struct.port), control_port(a_struct.control_port), 
          start_timeout(a_struct.start_timeout), stop_timeout(a_struct.stop_timeout), pid(a_struct.pid), 
          counter(a_struct.counter.load()), active(a_struct.active), leak_protected(a_struct.leak_protected)
      { }
    };

    subprocesses_t(size_t a_maxmem_leak_protection, size_t a_maxmem);
    ~subprocesses_t();
    int                     append(const char*                                        a_name,
                                   std::vector<std::string>                           a_argv1,
                                   std::vector<std::string>                           a_argv2,
                                   std::vector<std::pair<std::string, std::string> >  a_envv,
                                   std::vector<std::string>                           a_node_options,
                                   const char*                                        a_user,
                                   const char*                                        a_password,
                                   int                                                a_port1,
                                   int                                                a_port2,
                                   int                                                a_control_port1,
                                   int                                                a_control_port2,
                                   int                                                a_start_timeout,
                                   int                                                a_stop_timeout,
                                   bool                                               a_leak_protected);
    void                    run();
    void                    stop();
    int                     get_process(SUBPROCESS_LEAK_PROTECTED a_lp = SLP_NO_MATTER);
    subprocesses_state_t&   get_process_info(int a_index);
    void                    release_process(int a_index);
    bool                    is_active(int a_procNum);
    void                    restart();
    request_processing_information_t get_request_processing_information(std::string a_path, std::string a_context);
  private:
    void                    _stop_process(int a_index, bool a_hide_errors, bool a_force = false); // no exceptions
    void                    _restart_process(int a_index, bool a_restart);
    void                    _observer(int a_first_procindex, int a_second_procindex);
    void                    _send_close(int a_port, int a_timeout);
    void                    _send_environment(int a_port, int a_timeout);
    pid_t                   _send_require_subpid(int a_port, int a_timeout);

    size_t                              _maxmem_leak_protection;
    size_t                              _maxmem;
    bool                                _stop;
    std::vector<bool>                   _restart_flags;
    std::vector<std::thread>            _observers;
    std::vector<subprocesses_state_t>   _processes;
    std::mutex                          _mutex;
    std::condition_variable             _condition;
  };

  typedef std::shared_ptr<subprocesses_t> subprocesses_sptr_t;

}
